@@ -5,24 +5,55 @@ from argparse import ArgumentParser, ArgumentError, SUPPRESS, _SubParsersAction
 from argparse import OPTIONAL, ZERO_OR_MORE, ONE_OR_MORE, REMAINDER, PARSER
 from argparse import _get_action_name, _
 
+_num_consumed_args = {}
+
 
 def action_is_satisfied(action):
     ''' Returns False if the parse would raise an error if no more arguments are given to this action, True otherwise.
     '''
-    num_consumed_args = getattr(action, 'num_consumed_args', 0)
+    num_consumed_args = _num_consumed_args.get(action, 0)
 
-    if action.nargs == ONE_OR_MORE and num_consumed_args < 1:
+    if action.nargs in [OPTIONAL, ZERO_OR_MORE, REMAINDER]:
+        return True
+    if action.nargs == ONE_OR_MORE:
+        return num_consumed_args >= 1
+    if action.nargs == PARSER:
+        # Not sure what this should be, but this previously always returned False
+        # so at least this won't break anything that wasn't already broken.
         return False
-    elif action.nargs is None:
+    if action.nargs is None:
         return num_consumed_args == 1
-    else:
-        return num_consumed_args == action.nargs
+
+    assert isinstance(action.nargs, int), 'failed to handle a possible nargs value: %r' % action.nargs
+    return num_consumed_args == action.nargs
 
 
 def action_is_open(action):
     ''' Returns True if action could consume more arguments (i.e., its pattern is open).
     '''
-    return action.nargs in [ZERO_OR_MORE, ONE_OR_MORE, PARSER, REMAINDER]
+    num_consumed_args = _num_consumed_args.get(action, 0)
+
+    if action.nargs in [ZERO_OR_MORE, ONE_OR_MORE, PARSER, REMAINDER]:
+        return True
+    if action.nargs == OPTIONAL or action.nargs is None:
+        return num_consumed_args == 0
+
+    assert isinstance(action.nargs, int), 'failed to handle a possible nargs value: %r' % action.nargs
+    return num_consumed_args < action.nargs
+
+
+def action_is_greedy(action, isoptional=False):
+    ''' Returns True if action will necessarily consume the next argument.
+    isoptional indicates whether the argument is an optional (starts with -).
+    '''
+    num_consumed_args = _num_consumed_args.get(action, 0)
+
+    if action.option_strings:
+        if not isoptional and not action_is_satisfied(action):
+            return True
+        return action.nargs == REMAINDER
+    else:
+        return action.nargs == REMAINDER and num_consumed_args >= 1
 
 
 class IntrospectiveArgumentParser(ArgumentParser):
@@ -31,6 +62,8 @@ class IntrospectiveArgumentParser(ArgumentParser):
     '''
 
     def _parse_known_args(self, arg_strings, namespace):
+        _num_consumed_args.clear()  # Added by argcomplete
+        self._argcomplete_namespace = namespace
         self.active_actions = []  # Added by argcomplete
         # replace arg strings that are file references
         if self.fromfile_prefix_chars is not None:
@@ -39,6 +72,7 @@ class IntrospectiveArgumentParser(ArgumentParser):
         # map all mutually exclusive arguments to the other arguments
         # they can't occur with
         action_conflicts = {}
+        self._action_conflicts = action_conflicts  # Added by argcomplete
         for mutex_group in self._mutually_exclusive_groups:
             group_actions = mutex_group._group_actions
             for i, mutex_action in enumerate(mutex_group._group_actions):
@@ -77,6 +111,7 @@ class IntrospectiveArgumentParser(ArgumentParser):
         # converts arg strings to the appropriate and then takes the action
         seen_actions = set()
         seen_non_default_actions = set()
+        self._seen_non_default_actions = seen_non_default_actions  # Added by argcomplete
 
         def take_action(action, argument_strings, option_string=None):
             seen_actions.add(action)
@@ -95,10 +130,21 @@ class IntrospectiveArgumentParser(ArgumentParser):
 
             # take the action if we didn't receive a SUPPRESS value
             # (e.g. from a default)
-            # argcomplete: we should walk through sub-commands.
             if argument_values is not SUPPRESS \
                     or isinstance(action, _SubParsersAction):
-                action(self, namespace, argument_values, option_string)
+                try:
+                    action(self, namespace, argument_values, option_string)
+                except:
+                    # Begin added by argcomplete
+                    # When a subparser action is taken and fails due to incomplete arguments, it does not merge the
+                    # contents of its parsed namespace into the parent namespace. Do that here to allow completers to
+                    # access the partially parsed arguments for the subparser.
+                    if isinstance(action, _SubParsersAction):
+                        subnamespace = action._name_parser_map[argument_values[0]]._argcomplete_namespace
+                        for key, value in vars(subnamespace).items():
+                            setattr(namespace, key, value)
+                    # End added by argcomplete
+                    raise
 
         # function to convert arg_strings into an optional action
         def consume_optional(start_index):
@@ -161,7 +207,7 @@ class IntrospectiveArgumentParser(ArgumentParser):
                     start = start_index + 1
                     selected_patterns = arg_strings_pattern[start:]
                     self.active_actions = [action]  # Added by argcomplete
-                    action.num_consumed_args = 0  # Added by argcomplete
+                    _num_consumed_args[action] = 0  # Added by argcomplete
                     arg_count = match_argument(action, selected_patterns)
                     stop = start + arg_count
                     args = arg_strings[start:stop]
@@ -169,11 +215,9 @@ class IntrospectiveArgumentParser(ArgumentParser):
                     # Begin added by argcomplete
                     # If the pattern is not open (e.g. no + at the end), remove the action from active actions (since
                     # it wouldn't be able to consume any more args)
+                    _num_consumed_args[action] = len(args)
                     if not action_is_open(action):
                         self.active_actions.remove(action)
-                    elif action.nargs == OPTIONAL and len(args) == 1:
-                        self.active_actions.remove(action)
-                    action.num_consumed_args = len(args)
                     # End added by argcomplete
 
                     action_tuples.append((action, args, option_string))
@@ -199,14 +243,12 @@ class IntrospectiveArgumentParser(ArgumentParser):
 
             # slice off the appropriate arg strings for each Positional
             # and add the Positional and its args to the list
+            for action, arg_count in zip(positionals, arg_counts):  # Added by argcomplete
+                self.active_actions.append(action)  # Added by argcomplete
             for action, arg_count in zip(positionals, arg_counts):
-                if arg_count > 0:  # Added by argcomplete
-                    self.active_actions = [action]  # Added by argcomplete
-                else:  # Added by argcomplete
-                    self.active_actions.append(action)  # Added by argcomplete
                 args = arg_strings[start_index: start_index + arg_count]
                 start_index += arg_count
-                action.num_consumed_args = len(args)
+                _num_consumed_args[action] = len(args)   # Added by argcomplete
                 take_action(action, args)
 
             # slice off the Positionals that we just parsed and return the
